include "cysignals/signals.pxi"

from libcpp.vector cimport vector
from libcpp cimport bool
from libcpp.utility cimport pair
from sage.graphs.graph import Graph, DiGraph

# include because we don't want it to became external dependecy
include "parmap.py"
include "groupConnectivityNaive.py"

cdef extern from "group-connectivity.h" namespace "Ring":
  cdef cppclass Z4[T]:
    pass
  cdef cppclass Z2_2[T]:
    pass
  cdef cppclass Z5[T]:
    pass

cdef extern from "group-connectivity.h":
  ctypedef int EdgeId
  ctypedef int DirectedEdgeId
  ctypedef pair[DirectedEdgeId, DirectedEdgeId] TwoCut
  ctypedef vector[EdgeId] DoubleSubdivision

  cdef cppclass AbstractTester:
    void init(int, vector[EdgeId], vector[TwoCut], vector[DoubleSubdivision], vector[vector[DirectedEdgeId]])
    vector[size_t] getClasses()
    bool run()

  cdef cppclass Tester[T](AbstractTester):
    pass


def pathToEdges(G, path):
  """Transform list of vertices into list of edges."""
  ret = []
  u = path[0]
  for v in path[1:]:
    mul = 1
    if u > v: mul = -1
    ret.append(mul * G.edge_label(u, v))
    u = v
  return ret


def theOtherNeighbour(G, v, w):
  """Return neighbour of v in G not equal to w given v has degree 2."""
  N = G.neighbors(v)
  if N[0] == w:
    return N[1]
  else:
    return N[0]


def testGroupConnectivity(G, group = "Z4", getClasses = False,
                          useTwoCuts = True, useDoubleSubdivisions = True,
                          useSubgraphs = True, debug = False):
  cdef AbstractTester* tester = NULL
  if group == "Z4":
    tester = new Tester[Z4[int]]()
  elif group == "Z2_2":
    tester = new Tester[Z2_2[int]]()
  elif group == "Z5":
    tester = new Tester[Z5[int]]()

  assert(tester != NULL)

  if useSubgraphs and group in ["Z4", "Z2_2"] and not getClasses:
    for v in G.vertex_iterator():
      degs = sorted(G.degree(G.neighbors(v)))
      if degs[-1] <= 2:
        return False
      if len(degs) == 3 and degs[0] <= 2 and degs[1] <= 2:
        N = [ theOtherNeighbour(G, w, v) for w in G.neighbors(v) if G.degree(w) <= 2 ]
        if G.degree(N) == [ 2, 2 ]:
          return False
    for (u, v, _) in G.edge_iterator():
      if sorted(G.degree(G.neighbors(u))) == [ 2, 2, 3 ] and \
         sorted(G.degree(G.neighbors(v))) == [ 2, 2, 3 ]:
        return False

  if useTwoCuts:
    G.relabel()

  i = 1
  for (u, v, _) in G.edge_iterator():
    G.set_edge_label(u, v, i)
    i += 1

  spanningTree = Graph(G.min_spanning_tree(
    weight_function = lambda e: min(G.degree([e[0], e[1]]))
  ))
  elemCycles = [
    [ l ] + pathToEdges(G, spanningTree.shortest_path(u, v))
    for (u, v, l) in set(G.edge_iterator()) - set(spanningTree.edge_iterator())
  ]

  spanningTreeEdges = [ l for (_, _, l) in spanningTree.edge_iterator() ]
  availableEdges = set(spanningTreeEdges)

  doubleSubdivisions = []
  if useDoubleSubdivisions:
    for (u, v, l) in G.edges():
      if G.degree([u, v]) != [2, 2]:
        continue
      edges = set([ l for (_, _, l) in G.edges_incident([u, v]) ])
      if edges.issubset(availableEdges):
        doubleSubdivisions.append(edges)
        availableEdges -= edges

  twoCuts = []
  if useTwoCuts:
    for v in [ v for v in G.vertices() if G.degree(v) == 2 ]:
      e, f = G.edges_incident(v)
      el = e[2]
      fl = f[2]
      if el not in availableEdges or fl not in availableEdges:
        continue
      availableEdges.remove(el)
      availableEdges.remove(fl)

      w1 = e[0] if e[0] != v else e[1]
      w2 = f[0] if f[0] != v else f[1]

      if w1 > v:
        el *= -1
      if w2 < v:
        fl *= -1

      if debug:
        print "twoCut> ", v, w1, w2, el, fl, (w1 > v), (w2 < v)

      twoCuts.append((el, fl))

  if debug:
    print "Spanning Tree: ", spanningTreeEdges
    print "Elementary cycles: ", elemCycles
    print "Two cuts: ", twoCuts
    print "Double subdivided edges: ", doubleSubdivisions

  tester.init(G.num_edges(), spanningTreeEdges, twoCuts, doubleSubdivisions, elemCycles)
  sig_on()
  ret = tester.run()
  sig_off()

  if getClasses:
    classes = tester.getClasses()
    del tester
    return (ret, classes)
  else:
    del tester
    return ret


def subdivisionIterator(G, edges = None):
  if edges is None:
    edges = G.edges()

  if len(edges) == 0:
    yield G
    return

  for i in range(0, 3):
    H = G.copy()
    H.subdivide_edge(edges[0], i)
    for L in subdivisionIterator(H, edges[1:]):
      yield L


def guidedSubdivisionIterator(G, edges = None, guide = lambda H: (True, None), context = None):
  class DeadEnd(Exception):
    pass

  if edges is None:
    edges = G.edges()

  def impl(G, edges):
    if len(edges) == 0:
      if context is None:
        cont, val = guide(G)
      else:
        cont, val = guide(G, context)
      yield val
      if not cont:
        raise DeadEnd()
      return

    for i in range(0, 3):
      H = G.copy()
      H.subdivide_edge(edges[0], i)
      try:
        for L in impl(H, edges[1:]):
          yield L
      except DeadEnd:
        if i == 0:
          raise
        else:
          break

  try:
    for L in impl(G, edges):
      yield L
  except DeadEnd:
    pass


def generateSubdivisions(file, G, edges = None,
                         function = lambda H: H.graph6_string(),
                         parallel = 0, chunksize = 1, in_flight = None):
  f = open(file, 'w+')
  f.write("# Subdivisions of edges '%s' of graph '%s' (%s)\n" %
          (edges, G.graph6_string(), G))

  if parallel == 0:
    iterator = map(function, subdivisionIterator(G, edges))
  else:
    iterator = parmap(function, subdivisionIterator(G, edges),
                      nprocs = parallel, chunksize = chunksize,
                      chunks_in_flight = in_flight)

  for line in iterator:
    f.write(line)
    f.write("\n")

  f.close()

