# based on code written by klaus se from
# http://stackoverflow.com/a/16071616
# but now almost completly rewritten

import multiprocessing
from itertools import chain

def chunkme(X, chunksize):
  """Return items generated by X in chunks (lists) of size chunksize.

  EXAMPLE:

  >>> list(chunkme(range(20), 7))
  [[0, 1, 2, 3, 4, 5, 6], [7, 8, 9, 10, 11, 12, 13], [14, 15, 16, 17, 18, 19]]
  >>> list(chunkme(range(0), 2))
  []
  """
  chunk = []
  for x in X:
    chunk.append(x)
    if len(chunk) >= chunksize:
      yield chunk
      chunk = []
  if len(chunk):
    yield chunk

def producer_fun(X, q_in, cont, nprocs, chunksize):
  """parmap internal helper"""
  for i, x in enumerate(chunkme(X, chunksize)):
    cont.acquire()
    q_in.put((i, x))

  for _ in range(nprocs):
    q_in.put((None, None))

def worker_fun(f, q_in, q_out, multimap, chunksize):
  """parmap internal helper"""
  while True:
    i, chunk = q_in.get()
    if i is None:
      q_out.put((None, None))
      break

    ret = (f(x) for x in chunk)
    if multimap:
      ret = chain.from_iterable(ret)

    if chunksize is not None:
      for x in chunkme(ret, chunksize):
        q_out.put((i, x))
      q_out.put((i, None))
    else:
      q_out.put((i, list(ret)))


def parmap(f, X, nprocs = None, chunksize = 1, chunks_in_flight = None,
           in_order = True, multimap = False, out_chunksize = None):
    """Implements parallel version of map.

    Spawns producer and (multiple) worker threads. Works as generator
    calculating new values as needed (see below for fine tuning of ahead
    computation). Prevents memory exhaustion in case of fast workers
    and slow consumption of generated values.

    f -- function to map on each element of X
    X -- generator
    nprocs -- number of worker threads to spawn
    chunksize -- number of items retrieved by worker from queue at once
    chunks_in_flight -- maximal number of chunks being processed at once
    in_order -- if True, input sequence may be reordered during processing
    multimap -- if True, f is expected to return an Iterable and flattening is done
    out_chunksize -- size of chunks send from workers to consumer

    EXAMPLES:

    >>> list(parmap(lambda x: x+1, range(10)))
    [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    """
    if nprocs is None:
      nprocs = multiprocessing.cpu_count()

    if in_order:
      out_chunksize = None

    if out_chunksize is not None:
      out_chunksize = max(out_chunksize, 1)

    if chunks_in_flight is None:
      chunks_in_flight = 10 + 3 * nprocs

    chunks_in_flight = max(chunks_in_flight, nprocs + 1)

    cont   = multiprocessing.Semaphore(chunks_in_flight)
    q_in   = multiprocessing.Queue()
    q_out  = multiprocessing.Queue()

    proc = [ multiprocessing.Process(
        target = worker_fun,
        args = (f, q_in, q_out, multimap, out_chunksize)
      ) for _ in range(nprocs)]

    proc.append(multiprocessing.Process(
      target = producer_fun,
      args = (X, q_in, cont, nprocs, chunksize)
    ))

    for p in proc:
        p.daemon = True
        p.start()

    def get_chunk():
      ret = {}
      chunk_index = 0
      running_workers = nprocs

      while running_workers > 0:
        i, val = q_out.get()
        if i is None:
          running_workers -= 1
          continue

        if not in_order:
          if out_chunksize is None:
            cont.release()
          if val is None:
            cont.release()
            continue
          yield val
          continue

        ret[i] = val
        while chunk_index in ret:
          val = ret[chunk_index]
          chunk_index += 1
          cont.release()
          yield val

      for p in proc:
        p.join()

    return chain.from_iterable(get_chunk())


if __name__ == "__main__":
  import doctest
  (f, t) = doctest.testmod()
  print "%s: %i tests of %i failed." % (__file__, f, t)
  if f > 0:
    from sys import exit
    exit(1)

